class PowerwallError(Exception):
    def __init__(self, msg):
        super().__init__(msg)


class APIError(PowerwallError):
    def __init__(self, error):
        super().__init__("Powerwall api error: {}".format(error))


class APIChangedError(APIError):
    def __init__(self, response_class, json_response, added_attrs=[], removed_attrs=[]):
        self.json_response = json_response
        self.response_class = response_class
        self.added_attrs = added_attrs
        self.removed_attrs = removed_attrs

        msg = self._construct_msg()

        super().__init__(msg)

    def _construct_msg(self):
        msg = "It seems like the Powerwall API changed for '{}''".format(self.response_class)
        if len(self.added_attrs) > 0:
            if len(self.removed_attrs) > 0:
                msg = "{}: Attributes added: {}, removed attributes {}".format(msg, self.added_attrs, self.removed_attrs)
            else:
                msg = "{}: Some attributes where added to the response: {}".format(msg, self.added_attrs)
        elif len(self.removed_attrs) > 0:
            msg = "{}: Some attributes where removed from the response: {}".format(msg, self.removed_attrs)
        return msg


class PowerwallUnreachableError(PowerwallError):
    def __init__(self, reason=None):
        msg = "Powerwall is unreachable"
        self.reason = reason
        if reason is not None:
            msg = "{}: {}".format(msg, reason)
        super().__init__(msg)


class AccessDeniedError(PowerwallError):
    def __init__(self, resource, error=None):
        self.resource = resource
        self.error = error
        msg = "Access denied for resource {}".format(resource)
        if error is not None:
            msg = "{}: {}".format(msg, error)
        super().__init__(msg)



import json
import os
import unittest

import requests
import responses
from responses import GET, POST, Response, add

from tesla_powerwall import (
    AccessDeniedError,
    APIError,
    MetersAggregateResponse,
    MetersResponse,
    MeterType,
    Powerwall,
    PowerwallUnreachableError,
)

ENDPOINT = "https://1.1.1.1/api/"

METERS_RESPONSE = {
    "site": {
        "last_communication_time": "2020-04-09T05:50:38.989687241-07:00",
        "instant_power": -5347.455078125,
        "instant_reactive_power": -664.1942901611328,
        "instant_apparent_power": 5388.546173843879,
        "frequency": 49.99971389770508,
        "energy_exported": 5512641.122754764,
        "energy_imported": 9852397.795532543,
        "instant_average_voltage": 232.0439249674479,
        "instant_total_current": 0,
        "i_a_current": 0,
        "i_b_current": 0,
        "i_c_current": 0,
        "timeout": 1500000000,
    },
    "battery": {
        "last_communication_time": "2020-04-09T05:50:38.990165237-07:00",
        "instant_power": -10,
        "instant_reactive_power": 600,
        "instant_apparent_power": 600.0833275470999,
        "frequency": 49.995000000000005,
        "energy_exported": 4379890,
        "energy_imported": 5265110,
        "instant_average_voltage": 230.8,
        "instant_total_current": -0.4,
        "i_a_current": 0,
        "i_b_current": 0,
        "i_c_current": 0,
        "timeout": 1500000000,
    },
    "load": {
        "last_communication_time": "2020-04-09T05:50:38.974944676-07:00",
        "instant_power": 734.1549565813606,
        "instant_reactive_power": -469.988307011022,
        "instant_apparent_power": 871.7066645380579,
        "frequency": 49.99971389770508,
        "energy_exported": 0,
        "energy_imported": 24751111.13611111,
        "instant_average_voltage": 232.0439249674479,
        "instant_total_current": 3.1638620001982423,
        "i_a_current": 0,
        "i_b_current": 0,
        "i_c_current": 0,
        "timeout": 1500000000,
    },
    "solar": {
        "last_communication_time": "2020-04-09T05:50:38.974944676-07:00",
        "instant_power": 6099.032958984375,
        "instant_reactive_power": -422.27491760253906,
        "instant_apparent_power": 6113.633873631454,
        "frequency": 49.95012283325195,
        "energy_exported": 21296639.987777833,
        "energy_imported": 65.52444450131091,
        "instant_average_voltage": 232.1537322998047,
        "instant_total_current": 0,
        "i_a_current": 0,
        "i_b_current": 0,
        "i_c_current": 0,
        "timeout": 1500000000,
    },
}

SITE_INFO_RESPONSE_WITHOUT_OPTIONS = {
    "max_site_meter_power_kW": 1000000000,
    "min_site_meter_power_kW": -1000000000,
    "nominal_system_energy_kWh": 13.5,
    "nominal_system_power_kW": 5,
    "max_system_energy_kWh": 0,
    "max_system_power_kW": 0,
    "site_name": "Test",
    "timezone": "Europe/Rome",
    "grid_code": "50Hz_230V_1_CEI-021:2016_Italy",
    "grid_voltage_setting": 230,
    "grid_freq_setting": 50,
    "grid_phase_setting": "Single",
    "country": "Italy",
    "state": "*",
    "region": "CEI-021",
}
SITE_INFO_RESPONSE_WITH_OPTIONS = {
    "max_site_meter_power_kW": 1000000000,
    "min_site_meter_power_kW": -1000000000,
    "nominal_system_energy_kWh": 13.5,
    "nominal_system_power_kW": 10,
    "max_system_energy_kWh": 0,
    "max_system_power_kW": 0,
    "site_name": "Test",
    "timezone": "Europe/Berlin",
    "grid_code": "50Hz_230V_1_VDE4105:2011_Germany",
    "grid_voltage_setting": 230,
    "grid_freq_setting": 50,
    "grid_phase_setting": "Single",
    "country": "Germany",
    "state": "*",
    "distributor": "*",
    "utility": "*",
    "retailer": "*",
    "region": "VDE4105",
}


class TestPowerWall(unittest.TestCase):
    def setUp(self):
        self.powerwall = Powerwall(ENDPOINT)

    def test_endpoint_setup(self):
        test_endpoint_1 = "1.1.1.1"
        pw = Powerwall(test_endpoint_1)
        self.assertEqual(pw._endpoint, f"https://{test_endpoint_1}/api/")

        test_endpoint_2 = "http://1.1.1.1"
        pw = Powerwall(test_endpoint_2)
        self.assertEqual(pw._endpoint, f"https://1.1.1.1/api/")

        test_endpoint_3 = "https://1.1.1.1/api/"
        pw = Powerwall(test_endpoint_3)
        self.assertEqual(pw._endpoint, test_endpoint_3)

    @responses.activate
    def test_get_charge(self):
        add(
            Response(
                GET, url=f"{ENDPOINT}system_status/soe", json={"percentage": 53.123423}
            )
        )
        self.assertEqual(self.powerwall.get_charge(), 53)
        self.assertEqual(self.powerwall.get_charge(rounded=False), 53.123423)

    @responses.activate
    def test_process_response(self):
        res = requests.Response()
        res.request = requests.Request(method="GET", url=f"{ENDPOINT}test").prepare()
        res.status_code = 401
        with self.assertRaises(AccessDeniedError):
            self.powerwall._process_response(res)

        res.status_code = 502
        with self.assertRaises(PowerwallUnreachableError):
            self.powerwall._process_response(res)

        res.status_code = 200
        res._content = b'{"error": "test_error"}'
        with self.assertRaises(APIError):
            self.powerwall._process_response(res)

        res._content = b'{"response": "ok"}'
        self.assertEqual(self.powerwall._process_response(res), {"response": "ok"})

    @responses.activate
    def test_get(self):
        add(Response(GET, url=f"{ENDPOINT}test_get", json={"test_get": True}))

        self.assertEqual(self.powerwall._get("test_get"), {"test_get": True})

    @responses.activate
    def test_post(self):
        def post_callback(request):
            resp_body = {"test_post": True}
            headers = {}
            return (200, headers, json.dumps(resp_body))

        responses.add_callback(
            responses.POST,
            url=f"{ENDPOINT}test_post",
            callback=post_callback,
            content_type="application/json",
        )

        resp = self.powerwall._post("test_post", {"test": True})

        self.assertIsInstance(resp, dict)
        self.assertEqual(resp, {"test_post": True})

    @responses.activate
    def test_meters(self):
        add(
            Response(
                responses.GET, url=f"{ENDPOINT}meters/aggregates", json=METERS_RESPONSE
            )
        )
        meters = self.powerwall.get_meters()
        self.assertIsInstance(meters, MetersAggregateResponse)

        self.assertIsInstance(meters.site, MetersResponse)
        self.assertIsInstance(meters.solar, MetersResponse)
        self.assertIsInstance(meters.battery, MetersResponse)
        self.assertIsInstance(meters.load, MetersResponse)

    @responses.activate
    def test_is_sending(self):
        add(
            Response(
                responses.GET, url=f"{ENDPOINT}meters/aggregates", json=METERS_RESPONSE
            )
        )
        meters = self.powerwall.get_meters()
        self.assertEqual(meters.solar.is_sending_to(), False)
        self.assertEqual(meters.solar.is_active(), True)
        self.assertEqual(meters.solar.is_drawing_from(), True)
        self.assertEqual(meters.site.is_sending_to(), True)
        self.assertEqual(meters.load.is_sending_to(), True)
        self.assertEqual(meters.load.is_drawing_from(), False)
        self.assertEqual(meters.load.is_active(), True)

    @responses.activate
    def test_optional_json_attrs(self):
        add(
            Response(
                responses.GET,
                url=f"{ENDPOINT}site_info",
                json=SITE_INFO_RESPONSE_WITHOUT_OPTIONS,
            )
        )

        add(
            Response(
                responses.GET,
                url=f"{ENDPOINT}site_info",
                json=SITE_INFO_RESPONSE_WITH_OPTIONS,
            )
        )

        site_info = self.powerwall.get_site_info()
        self.assertEqual(site_info.has_optional_attrs_set(), False)
        self.assertEqual(site_info.has_key("utility"), False)
        self.assertEqual(site_info.grid_voltage_setting, 230)
        self.assertEqual(site_info.has_optional_attrs(), True)

        site_info = self.powerwall.get_site_info()
        self.assertEqual(site_info.has_optional_attrs_set(), True)
        self.assertEqual(site_info.has_key("utility"), True)
        self.assertEqual(site_info.retailer, "*")



from json.decoder import JSONDecodeError
from typing import List, Union
from urllib.parse import urljoin, urlparse, urlsplit, urlunparse, urlunsplit
from packaging import version
from contextlib import contextmanager

import requests
from requests import Session
from urllib3 import disable_warnings
from urllib3.exceptions import InsecureRequestWarning

from .const import (DEFAULT_KW_ROUND_PERSICION, SUPPORTED_OPERATION_MODES,
                    SUPPORTED_POWERWALL_VERSIONS, DeviceType, GridState,
                    GridStatus, LineStatus, MeterType, OperationMode, User,
                    Version)
from .error import AccessDeniedError, APIError, PowerwallUnreachableError, APIChangedError, PowerwallError
from .helpers import convert_to_kw
from .responses import (CustomerRegistrationResponse, ListPowerwallsResponse,
                        LoginResponse, MeterDetailsResponse,
                        MetersAggregateResponse, MetersResponse,
                        PowerwallsStatusResponse, PowerwallStatusResponse,
                        SiteInfoResponse, SitemasterResponse, SolarsResponse,
                        UpdateStatusResponse)

VERSION = "0.2.9"


class Powerwall(object):
    def __init__(
        self,
        endpoint: str,
        timeout: int = 10,
        http_session: requests.Session = None,
        verify_ssl: bool = False,
        disable_insecure_warning: bool = True,
        dont_validate_response: bool = True,
        pin_version: str = None,
    ):

        if disable_insecure_warning:
            disable_warnings(InsecureRequestWarning)

        if endpoint.startswith("https"):
            self._endpoint = endpoint
        elif endpoint.startswith("http"):
            self._endpoint = endpoint.replace("http", "https")
        else:
            # Use str.format instead of f'strings to be backwards compatible
            self._endpoint = "https://{}".format(endpoint)

        if not self._endpoint.endswith("api") and not self._endpoint.endswith("/"):
            self._endpoint += "/api/"
        elif self._endpoint.endswith("api"):
            self._endpoint += "/"
        elif self._endpoint.endswith("/") and not self._endpoint.endswith("api/"):
            self._endpoint += "api/"

        self._timeout = timeout
        self._http_session = http_session if http_session else Session()
        self._http_session.verify = verify_ssl
        self._dont_validate_response = dont_validate_response

        self._token = None

        self.pin_version(pin_version)

    def _process_response(self, response: str) -> dict:
        if response.status_code == 404:
            raise APIError("The url {} returned error 404".format(response.request.path_url))

        if response.status_code == 401 or response.status_code == 403:
            response_json = None
            try:
                response_json = response.json()
            except Exception:
                raise AccessDeniedError(response.request.path_url)
            else:
                raise AccessDeniedError(
                    response.request.path_url, response_json["error"]
                )

        if response.status_code == 502:
            raise PowerwallUnreachableError()

        try:
            response_json = response.json()
        except JSONDecodeError:
            raise APIError("Error while decoding json of response: {}".format(response.text))

        if response_json is None:
            return {}

        if "error" in response_json:
            raise APIError(response_json["error"])

        return response_json

    def _get(
        self, path: str, needs_authentication: bool = False, headers: dict = {}
    ) -> dict:
        if needs_authentication and not self.is_authenticated():
            raise APIError("Authentication required to access {}".format(path))

        try:
            response = self._http_session.get(
                url=urljoin(self._endpoint, path),
                timeout=self._timeout,
                headers=headers,
            )
        except (requests.exceptions.ConnectionError, requests.exceptions.ReadTimeout) as e:
            raise PowerwallUnreachableError(e)

        return self._process_response(response)

    def _post(
        self,
        path: str,
        payload: dict,
        needs_authentication: bool = False,
        headers: dict = {},
    ) -> dict:
        if needs_authentication and not self.is_authenticated():
            raise APIError("Authentication required to access {}".format(path))

        try:
            response = self._http_session.post(
                url=urljoin(self._endpoint, path),
                data=payload,
                timeout=self._timeout,
                headers=headers,
            )
        except (requests.exceptions.ConnectionError, requests.exceptions.ReadTimeout) as e:
            raise PowerwallUnreachableError(e)

        return self._process_response(response)

    def is_authenticated(self) -> bool:
        return "AuthCookie" in self._http_session.cookies.keys()

    def login_as(
        self,
        user: Union[User, str],
        email: str,
        password: str,
        force_sm_off: bool = False,
    ) -> LoginResponse:
        if isinstance(user, User):
            user = user.value

        # force_sm_off is referred to as 'shouldForceLogin' in the web source code
        response = self._post(
            "login/Basic",
            {
                "username": user,
                "email": email,
                "password": password,
                "force_sm_off": force_sm_off,
            },
        )

        # The api returns an auth cookie which is automatically set
        # so there is no need to further process the response

        return LoginResponse(response, no_check=self._dont_validate_response)

    def login(
        self, email: str, password: str, force_sm_off: bool = False
    ) -> LoginResponse:
        return self.login_as(User.CUSTOMER, email, password, force_sm_off)

    def logout(self):
        if not self.is_authenticated():
            raise APIError("Must be logged in to log out")
        # The api unsets the auth cookie and the token is invalidated
        self._get("logout", True)

    def run(self):
        self._get("sitemaster/run", True)

    def stop(self):
        self._get("sitemaster/stop", True)

    def get_charge(self, rounded: bool = True) -> Union[float, int]:
        """Returns current charge of powerwall"""
        charge = self._get("system_status/soe")["percentage"]
        if rounded:
            return round(charge)
        else:
            return charge

    def get_sitemaster(self) -> SitemasterResponse:
        return SitemasterResponse(
            self._get("sitemaster"), no_check=self._dont_validate_response
        )

    def get_meters(self) -> MetersAggregateResponse:
        """Returns the different meters in a MetersAggregateResponse"""
        return MetersAggregateResponse(
            self._get("meters/aggregates"), no_check=self._dont_validate_response
        )

    def get_meter_details(
        self, meter: MeterType
    ) -> Union[List[MeterDetailsResponse], None]:
        """Returns details about a specific meter. 
        If their are no details available for a meter None is returned."""
        resp = self._get("meters/{}".format(meter.value))
        if isinstance(resp, list):
            return [
                MeterDetailsResponse(item, no_check=self._dont_validate_response)
                for item in resp
            ]
        elif isinstance(resp, dict):
            return None

    def get_meter_readings(self) -> dict:
        return self._get("meter/readings", True)

    def get_grid_status(self) -> GridStatus:
        """Returns the current grid status."""
        return GridStatus(self._get("system_status/grid_status")["grid_status"])

    def get_grid_services_active(self) -> bool:
        return self._get("system_status/grid_status")["grid_services_active"]

    def get_grid_codes(self) -> list:
        """Returns all available grid codes"""
        return self._get("site_info/grid_codes", needs_authentication=True)

    def get_site_info(self) -> SiteInfoResponse:
        """Returns information about the powerwall site"""
        return SiteInfoResponse(
            self._get("site_info"), no_check=self._dont_validate_response
        )

    def set_site_name(self, site_name: str):
        return self._post("site_info/site_name", {"site_name": site_name}, True)

    def get_status(self) -> PowerwallStatusResponse:
        return PowerwallStatusResponse(
            self._get("status"), no_check=self._dont_validate_response
        )

    def get_powerwalls_status(self) -> PowerwallsStatusResponse:
        return PowerwallsStatusResponse(
            self._get("powerwalls/status", True), no_check=self._dont_validate_response
        )

    def get_device_type(self) -> DeviceType:
        """Returns the device type of the powerwall"""
        if self._pin_version is None or self._pin_version >= Version.v1_46_0.value:
            return self.get_status().device_type
        else:
            return DeviceType(self._get("device_type")["device_type"])

    def get_customer_registration(self) -> CustomerRegistrationResponse:
        return CustomerRegistrationResponse(
            self._get("customer/registration"), no_check=self._dont_validate_response
        )

    def get_powerwalls(self) -> ListPowerwallsResponse:
        """Returns powerwalls and status"""
        return ListPowerwallsResponse(
            self._get("powerwalls"), no_check=self._dont_validate_response
        )

    def get_serial_numbers(self) -> List[str]:
        return [
            powerwall.PackageSerialNumber
            for powerwall in self.get_powerwalls().powerwalls
        ]

    def get_operation_mode(self) -> OperationMode:
        return OperationMode(self._get("operation", True)["real_mode"])

    def get_backup_preserve_percentage(self) -> float:
        return self._get("operation", True)["backup_reserve_percentage"]

    # def set_mode_and_backup_preserve_percentage(self, mode, percentage):
    #     self._post("operation", {"mode": mode, "percentage": percentage})

    # def set_backup_preserve_percentage(self, percentage):
    #     self.set_mode_and_backup_preserve_percentage(self.mode, percentage)

    # def set_mode(self, mode):
    #     self.set_mode_and_backup_preserve_percentage(
    #         mode, self.backup_preserve_percentage)

    def get_networks(self) -> list:
        return self._get("networks")

    def get_phase_usage(self):
        return self._get("powerwalls/phase_usages", needs_authentication=True)

    def set_run_for_commissioning(self):
        self._post("sitemaster/run_for_commissioning", True)

    def get_solars(self) -> List[SolarsResponse]:
        solars = self._get("solars", needs_authentication=True)
        return [
            SolarsResponse(solar, no_check=self._dont_validate_response)
            for solar in solars
        ]

    def get_vin(self) -> str:
        return self._get("config", needs_authentication=True)["vin"]

    def get_logs(self) -> str:
        return self._get("getlogs", needs_authentication=True)

    def get_meters_info(self) -> list:
        return self._get("meters", needs_authentication=True)

    def get_installer_info(self) -> dict:
        return self._get("installer", needs_authentication=True)

    def get_solar_brands(self) -> List[str]:
        return self._get("solars/brands", needs_authentication=True)

    def get_version(self) -> str:
        return self.get_status().version

    def get_git_hash(self) -> str:
        return self.get_status().git_hash

    def get_update_status(self) -> UpdateStatusResponse:
        return UpdateStatusResponse(self._get("system/update/status", needs_authentication=True))

    def is_sending_to(self, meter: MeterType, rounded=True) -> bool:
        """Wrapper method for is_sending_to"""
        return self.get_meters().get(meter).is_sending_to()

    def is_drawing_from(self, meter: MeterType) -> bool:
        """Wrapper method for is_drawing_from"""
        return self.get_meters().get(meter).is_drawing_from()

    def is_active(self, meter: MeterType) -> bool:
        """Wrapper method for is_active()"""
        return self.get_meters().get(meter).is_active()

    def get_power(self, meter: MeterType) -> bool:
        return self.get_meters().get(meter).get_power()

    def is_powerwall_version_supported(self) -> bool:
        return self.get_version() in SUPPORTED_POWERWALL_VERSIONS

    def set_dont_validate_response(self, value):
        self._dont_validate_response = value

    @contextmanager
    def no_verify(self):
        self.set_dont_validate_response(True)
        try:
            yield
        finally:
            self.set_dont_validate_response(False)


    def detect_and_pin_version(self) -> str:
        """Does an unferified request to get powerwall version and pins it"""
        with self.no_verify():
            status = self.get_status()
            if status.has_key("version"):
                version = status.get("version")
            else:
                raise APIError(
                    "Could not detect version because the status response does not return the version"
                )

        self.pin_version(version)

    def pin_version(self, vers: Union[str, version.Version, Version]):
        if vers is None:
            self._pin_version = None
        else:
            if isinstance(vers, str):
                self._pin_version = version.parse(vers)
            elif isinstance(vers, Version):
                self._pin_version = vers.value
            else:
                self._pin_version = vers

    def get_pinned_version(self) -> version.Version:
        return self._pin_version




from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="tesla_powerwall",
    version='0.2.9',
    description="API for Tesla Powerwall",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jrester/tesla_powerwall",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3"
    ],
    install_requires=[
        "requests>=2.22.0",
        "packaging>=20.0"
    ]
)



def convert_to_kw(value: float, precision: int = 1) -> float:
    """Converts watt to kilowatt and rounds to precision"""
    # Don't round if precision is -1
    if precision == -1:
        return value / 1000
    else:
        return round(value / 1000, precision)
